/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

/**
 * Utility methods for buffers
 */
public class BufferUtils
{
    /**
     * Create a direct byte buffer with the given size
     * 
     * @param size The size of the buffer
     * @return The buffer
     */
    public static ByteBuffer createByteBuffer(int size)
    {
        return ByteBuffer.allocateDirect(size).
            order(ByteOrder.nativeOrder());
    }

    /**
     * Create a direct float buffer with the given size
     * 
     * @param size The size of the buffer
     * @return The buffer
     */
    public static FloatBuffer createFloatBuffer(int size)
    {
        return createByteBuffer(size * 4).asFloatBuffer();
    }

    /**
     * Create a direct int buffer with the given size
     * 
     * @param size The size of the buffer
     * @return The buffer
     */
    public static IntBuffer createIntBuffer(int size)
    {
        return createByteBuffer(size * 4).asIntBuffer();
    }
	
    /**
     * Create a direct byte buffer which has the contents
     * of the given array.
     * 
     * @param a The array
     * @return The buffer
     */
    public static ByteBuffer createDirectBuffer(byte a[])
    {
        ByteBuffer b = createByteBuffer(a.length);
        b.put(a);
        b.rewind();
        return b;
    }

    /**
     * Create a direct int buffer which has the contents
     * of the given array.
     * 
     * @param a The array
     * @return The buffer
     */
    public static IntBuffer createDirectBuffer(int a[])
    {
        IntBuffer b = createIntBuffer(a.length);
        b.put(a);
        b.rewind();
        return b;
    }

    /**
     * Create a direct float buffer which has the contents
     * of the given array.
     * 
     * @param a The array
     * @return The buffer
     */
    public static FloatBuffer createDirectBuffer(float a[])
    {
        FloatBuffer b = createFloatBuffer(a.length);
        b.put(a);
        b.rewind();
        return b;
    }

    
    
    
    
//    /**
//     * Returns a String representation of the given buffer
//     * 
//     * @param buffer The buffer
//     * @return A String representation of the buffer
//     */
//	public static String toString(FloatBuffer buffer)
//	{
//		StringBuilder sb = new StringBuilder();
//		sb.append("(");
//		for (int i=0; i<buffer.capacity(); i++)
//		{
//			sb.append(buffer.get(i));
//			if (i<buffer.capacity()-1)
//			{
//				sb.append(", ");
//			}
//		}
//		sb.append(")");
//		return sb.toString();
//	}
//	
//	/**
//	 * Converts the given String into a 0-terminated, direct
//	 * byte buffer.
//	 * 
//	 * @param s The input String
//	 * @return The byte buffer
//	 */
//    public static ByteBuffer toByteBuffer(String s)
//    {
//        byte bytes[] = s.getBytes();
//        ByteBuffer buffer = createByteBuffer(bytes.length+1);
//        buffer.put(bytes);
//        buffer.put((byte)0);
//        buffer.rewind();
//        return buffer;
//    }
 
    
    /**
     * Private constructor to prevent instantiation
     */
    private BufferUtils()
    {
    }
    
}