/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.awt.Point;

import javax.swing.JComponent;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Tuple2f;
import javax.vecmath.Tuple3f;

import de.javagl.swogl.geometry.Geometry;
import de.javagl.swogl.geometry.GeometryUtils;

/**
 * This class summarizes information about the intersection between
 * a ray and a {@link SwoglComponent}. Instances of this class may
 * be obtained from 
 * {@link SwoglContainer#computeIntersectionInfos(java.awt.Point)}.
 */
public final class IntersectionInfo
{
    /**
     * The SwoglComponent that was intersected
     */
    private final SwoglComponent swoglComponent;
    
    /**
     * The distance of the intersection from the ray origin
     */
    private final float distance;

    /**
     * The index of the intersected triangle of the geometry
     * of the SwoglComponent.
     */
    private final int triangleIndex;
    
    /**
     * The barycentric coordinates of the intersection.
     */
    private final Point3f baryCoords;
    
    /**
     * The {@link PickingInfo} for this IntersectionInfo
     */
    private PickingInfo pickingInfo = null;
    
    /**
     * The position where the ray intersected the SwoglComponent.
     * Stored in the local coordinate system of the SwoglComponent.
     */
    private Point3f intersectionPosition = null;

    /**
     * Creates a new IntersectionInfo with the given parameters.
     * 
     * @param swoglComponent The intersected SwoglComponent
     * @param triangleIndex The index of the intersected triangle
     * @param distance The distance of the intersection
     * @param baryCoords The barycentric coordinates of the intersection. 
     */
    IntersectionInfo(SwoglComponent swoglComponent, float distance, 
        int triangleIndex, Point3f baryCoords)
    {
        this.swoglComponent = swoglComponent;
        this.triangleIndex = triangleIndex;
        this.distance = distance;
        this.baryCoords = new Point3f(baryCoords);
    }

    /**
     * Returns the SwoglComponent that was intersected
     * 
     * @return The SwoglComponent that was intersected
     */
    public SwoglComponent getSwoglComponent()
    {
        return swoglComponent;
    }

    /**
     * Returns the distance of the intersection from the ray origin
     * 
     * @return The distance of the intersection from the ray origin
     */
    public float getDistance()
    {
        return distance;
    }

    /**
     * Returns the index of the intersected triangle of the geometry
     * of the SwoglComponent.
     * 
     * @return The index of the intersected triangle
     */
    int getTriangleIndex()
    {
        return triangleIndex;
    }

    /**
     * Returns a copy of the barycentric coordinates where the
     * triangle was intersected.
     * 
     * @return The barycentric coordinates of the intersection.
     */
    public Point3f getBaryCoords()
    {
        return new Point3f(baryCoords);
    }
    
    /**
     * Compute the position where the intersection happened, in the
     * local coordinate system of the SwoglComponent, and return 
     * it as a Point. To convert this position into world coordinates,
     * it has to be transformed with the transformation matrix of 
     * the SwoglComponent.
     *  
     * @return The intersection position
     */
    public Point3f computeIntersectionPosition()
    {
        if (intersectionPosition == null)
        {
            intersectionPosition = new Point3f();
            
            // Obtain the positions of the vertices of the hit triangle
            Tuple3f t0 = new Point3f();
            Tuple3f t1 = new Point3f();
            Tuple3f t2 = new Point3f();
            Geometry geometry = swoglComponent.getGeometry();
            GeometryUtils.getVertices(
                geometry, triangleIndex, t0, t1, t2);
            
            // Compute the intersection position
            intersectionPosition.scaleAdd(
                baryCoords.x, t0, intersectionPosition);
            intersectionPosition.scaleAdd(
                baryCoords.y, t1, intersectionPosition);
            intersectionPosition.scaleAdd(
                baryCoords.z, t2, intersectionPosition);
        }
        return new Point3f(intersectionPosition);
    }
    
    /**
     * Returns the {@link PickingInfo} for the intersection described
     * by this IntersectionInfo
     * 
     * @return The PickingInfo for this intersection
     */
    public PickingInfo createPickingInfo()
    {
        if (pickingInfo == null)
        {
            // Obtain the texture coordinates of the hit triangle
            Tuple2f t0 = new TexCoord2f();
            Tuple2f t1 = new TexCoord2f();
            Tuple2f t2 = new TexCoord2f();
            Geometry geometry = swoglComponent.getGeometry();
            GeometryUtils.getTexCoords(
                geometry, triangleIndex, t0, t1, t2);
            
            // Compute the texture coordinates at the intersection position
            Tuple2f closestTexCoords = new TexCoord2f();
            closestTexCoords.scaleAdd(baryCoords.x, t0, closestTexCoords);
            closestTexCoords.scaleAdd(baryCoords.y, t1, closestTexCoords);
            closestTexCoords.scaleAdd(baryCoords.z, t2, closestTexCoords);
            
            // Create the PickingInfo 
            JComponent contents = swoglComponent.getComponent();
            int cx = (int)(contents.getWidth() * closestTexCoords.x);
            int cy = (int)(contents.getHeight() * closestTexCoords.y);
            pickingInfo = new PickingInfo(contents, new Point(cx, cy));
        }
        return pickingInfo;
    }
    

    @Override
    public String toString()
    {
        return "IntersectionInfo[" +
            "distance=" + distance + "," +
            "triangleIndex=" + triangleIndex + "," +
            "baryCoords=" + baryCoords + "," +
            "swoglComponent=" + swoglComponent + "]";
    }
    
}