/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.util.ServiceLoader;
import java.util.logging.Logger;

/**
 * This class is responsible for resolving services that may create 
 * {@link SwoglApplication SwoglApplications}. It will resolve 
 * services that implement the {@link SwoglApplicationFactory}
 * interface, and use the factory to create a new SwoglApplication.
 */
public class SwoglApplications
{
    /**
     * The logger used in this class
     */
    private static final Logger logger = 
        Logger.getLogger(SwoglApplications.class.getName());
    
    /**
     * The ServiceLoader for SwoglApplicationFactories
     */
    private static final ServiceLoader<SwoglApplicationFactory> 
        swoglApplicationFactories = 
            ServiceLoader.load(SwoglApplicationFactory.class);
    
    
    /**
     * Creates a new {@link SwoglApplication}. This method will
     * try to obtain a service that implements the 
     * {@link SwoglApplicationFactory} interface. This service
     * will be used to instantiate a SwoglApplication. The first 
     * SwoglApplication that can successfully be instantiated
     * will be returned.
     * 
     * @return A new SwoglApplication, or <code>null</code> if no
     * service for creating a SwoglApplication could be found.
     */
    public static SwoglApplication createSwoglApplication() 
    {
        for (SwoglApplicationFactory factory : swoglApplicationFactories) 
        {
            logger.config("Found factory "+factory);
        }
        
        // For testing:
        String preference = "";
        //preference = "JOGL";
        //preference = "LWJGL";
        //preference = "JavaGL";
        for (SwoglApplicationFactory factory : swoglApplicationFactories) 
        {
            String className = factory.getClass().getName();
            if (className.contains(preference))
            {
                try
                {
                    SwoglApplication swoglApplication = 
                        factory.createSwoglApplication();
                    logger.config("Created "+swoglApplication);
                    if (swoglApplication != null)
                    {
                        return swoglApplication;
                    }
                }
                catch (Exception t)
                {
                    logger.warning("Failed to instantiate SwoglApplication " +
                    	"using factory "+factory);
                }
            }
        }
        return null;
    }      
    
    /**
     * Private constructor to prevent instantiation
     */
    private SwoglApplications()
    {
    }
}
