/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

/**
 * Interface for all classes that may handle SwoglComponents. Instances
 * of classes implementing this interface may be passed to the
 * {@link SwoglContainer#create(java.awt.Component, SwoglComponentHandler, 
 *  PickingRayComputer)} method, and will be responsible for maintaining the
 * rendered representation of {@link SwoglComponent SwoglComponents}:
 * The SwoglComponents may be added to or removed from the SwoglContainer, 
 * which will cause the 
 * {@link SwoglComponentHandler#handleSwoglComponent(SwoglComponent)} and
 * {@link SwoglComponentHandler#releaseSwoglComponent(SwoglComponent)} 
 * methods to be called, respectively. The handler is responsible
 * for maintaining the internal rendering data structures associated with
 * the respective SwoglComponent.
 */
public interface SwoglComponentHandler
{
    /**
     * Handle the given SwoglComponent. This method will be called
     * by the SwoglContainer, when a new SwoglComponent is added.
     * This may eventually cause internal data structures for the
     * SwoglComponent to be built, so that the SwoglComponent may
     * be rendered. The method may be called from any thread. 
     * The implementor is responsible for making this method 
     * thread-safe. 
     *  
     * @param swoglComponent The SwoglComponent to handle
     */
    void handleSwoglComponent(SwoglComponent swoglComponent);

    /**
     * Release the given SwoglComponent. This method will be called
     * by the SwoglContainer, when a new SwoglComponent is removed.
     * This may eventually cause internal data structured for the
     * SwoglComponent to be cleaned up. The method may be called 
     * from any thread. The implementor is responsible for making 
     * this method thread-safe. 
     *  
     * @param swoglComponent The SwoglComponent to release
     */
    void releaseSwoglComponent(SwoglComponent swoglComponent);
    
    
    
}
