/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.geometry;


import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Tuple2f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Basic implementation of a {@link Geometry}. A DefaultGeometry 
 * simply creates a rectangular triangle mesh with a specified number of 
 * vertices. The default vertex coordinates will range from 
 * (-width/2, -height/2) to (width/2,height/2), and the texture
 * coordinates will be in [0,1]x[0,1]. <br />
 * <br />
 */ 
final class DefaultGeometry extends AbstractGeometry 
    implements SimpleGeometry
{
    /** 
     * The number of vertices in x direction
     */
    private final int verticesX;

    /**
     * The number of vertices in y direction
     */
    private final int verticesY;

    /**
     * Creates a new DefaultGeometry with the given size and 
     * number of vertices
     * 
     * @param width The size in x direction
     * @param height The size in y direction
     * @param verticesX The number of vertices in x direction
     * Will be clamped to be at least 2.
     * @param verticesY The number of vertices in y direction 
     * Will be clamped to be at least 2.
     */
    DefaultGeometry(
        int width, int height, int verticesX, int verticesY)
    {
        this.verticesX = Math.max(2, verticesX);
        this.verticesY = Math.max(2, verticesY);

        initVertices();
        initNormals();
        initTexCoords();
        initIndices();

        setupVertices(width, height);
        setupTexCoords();
        
        setVerticesModified(true);
        setTexCoordsModified(true);
    }

    @Override
    public int getNumPointsX()
    {
        return verticesX;
    }
    
    
    @Override
    public int getNumPointsY()
    {
        return verticesY;
    }

    @Override
    public void getVertex(int x, int y, Tuple3f vertex)
    {
        int index = x * verticesY + y;
        getVertex(index, vertex);
    }

    @Override
    public void setVertex(int x, int y, Tuple3f vertex)
    {
        int index = x * verticesY + y;
        setVertex(index, vertex);
    }

    @Override
    public void getTexCoord(int x, int y, Tuple2f texCoord)
    {
        int index = x * verticesY + y;
        getTexCoord(index, texCoord);
    }

    @Override
    public void setTexCoord(int x, int y, Tuple2f texCoord)
    {
        int index = x * verticesY + y;
        setTexCoord(index, texCoord);
    }

    @Override
    public void getNormal(int x, int y, Tuple3f normal)
    {
        int index = x * verticesY + y;
        getNormal(index, normal);
    }

    @Override
    public void setNormal(int x, int y, Tuple3f normal)
    {
        int index = x * verticesY + y;
        setNormal(index, normal);
    }

    /**
     * Initialize the indices, so that three consecutive indices
     * are the indices of the vertices of one triangle
     */
    private void initIndices()
    {
        for (int sx = 0; sx < verticesX - 1; sx++)
        {
            for (int sy = 0; sy < verticesY - 1; sy++)
            {
                int i0 = sy + sx * verticesY;
                int i1 = sy + (sx + 1) * verticesY;
                int i2 = (sy + 1) + (sx + 1) * verticesY;
                addTriangle(i0, i1, i2);

                i0 = sy + sx * verticesY;
                i1 = (sy + 1) + (sx + 1) * verticesY;
                i2 = (sy + 1) + sx * verticesY;
                addTriangle(i0, i1, i2);
            }
        }
    }
    
    /**
     * Initialize the vertices with vertices that are all (0,0,0)
     */
    private void initVertices()
    {
        Point3f vertex = new Point3f();
        for (int x = 0; x < verticesX; x++)
        {
            for (int y = 0; y < verticesY; y++)
            {
                addVertex(vertex);
            }
        }
    }

    /**
     * Initialize the normals with normals that are all (0,0,1)
     */
    private void initNormals()
    {
        Vector3f normal = new Vector3f(0,0,1);
        for (int x = 0; x < verticesX; x++)
        {
            for (int y = 0; y < verticesY; y++)
            {
                addNormal(normal);
            }
        }
    }
    
    /**
     * Initialize the texture coordinates for this
     * geometry, so that they initially are all (0,0)
     */
    private void initTexCoords()
    {
        TexCoord2f texCoord = new TexCoord2f();
        for (int sx = 0; sx < verticesX; sx++)
        {
            for (int sy = 0; sy < verticesY; sy++)
            {
                addTexCoord(texCoord);
            }
        }
    }
    
    
    /**
     * Set up the vertex positions according to the given
     * width and height, so that they form a plane mesh.
     * 
     * @param width The width for the mesh
     * @param height The height of the mesh
     */
    private void setupVertices(int width, int height)
    {
        int x = 0;
        int y = 0;
        float stepX = 1.0f / (verticesX - 1);
        float stepY = 1.0f / (verticesY - 1);
        Point3f vertex = new Point3f();
        for (int sx = 0; sx < verticesX; sx++)
        {
            for (int sy = 0; sy < verticesY; sy++)
            {
                float relX = sx * stepX;
                float relY = sy * stepY;
                
                float posX = -0.5f + relX;
                float posY = -0.5f + relY;

                int index = sx * verticesY + sy;

                vertex.x = x + posX * width;
                vertex.y = y + posY * height;
                vertex.z = 0;
                setVertex(index, vertex);
            }
        }
    }

    /**
     * Initialize the texture coordinates, so that they range
     * from (0,0) to (1,1). 
     */
    private void setupTexCoords()
    {
        float dx = 1.0f / (verticesX - 1);
        float dy = 1.0f / (verticesY - 1);
        TexCoord2f texCoord = new TexCoord2f();
        for (int sx = 0; sx < verticesX; sx++)
        {
            for (int sy = 0; sy < verticesY; sy++)
            {
                int index = sx * verticesY + sy;
                texCoord.x = sx * dx;
                texCoord.y = 1 - (sy * dy);
                setTexCoord(index, texCoord);
            }
        }
    }

    @Override
    public String toString()
    {
        return "DefaultGeometry[" +
        	"verticesX=" + verticesX + "," +
        	"verticesY=" + verticesY + "]";
    }

}
