/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.geometry;


import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Vector3f;

import de.javagl.swogl.geometry.AbstractGeometry;
import de.javagl.swogl.geometry.Geometry;

/**
 * Implementation of a {@link Geometry} that stores a rectangle
 * consisting of individual triangles (i.e. triangles that do 
 * not have common vertices).
 * The vertex coordinates will range from (-width/2, -height/2) 
 * to (width/2,height/2), and the texture coordinates will be 
 * in [0,1]x[0,1]. 
 */ 
public final class TriangleSetGeometry extends AbstractGeometry 
{
    /**
     * Creates a new TriangleSetGeometry with the given size and 
     * number of vertices
     * 
     * @param width The size in x direction
     * @param height The size in y direction
     * @param verticesX The number of grid vertices in x direction
     * Will be clamped to be at least 2.
     * @param verticesY The number of grid vertices in y direction 
     * Will be clamped to be at least 2.
     */
    public TriangleSetGeometry(
        int width, int height, int verticesX, int verticesY)
    {
        init(width, height, Math.max(2, verticesX), Math.max(2, verticesY));
        setVerticesModified(true);
        setTexCoordsModified(true);
    }

    /**
     * Create all vertices
     * 
     * @param width The size in x direction
     * @param height The size in y direction
     * @param verticesX The number of grid vertices in x direction
     * @param verticesY The number of grid vertices in y direction 
     */
    private void init(int width, int height, int verticesX, int verticesY)
    {
        for (int sx = 0; sx < verticesX - 1; sx++)
        {
            for (int sy = 0; sy < verticesY - 1; sy++)
            {
                int i0, i1, i2;

                i0 = getNumVertices();
                createVertex(sx  , sy, width, height, verticesX, verticesY  );
                i1 = getNumVertices();
                createVertex(sx+1, sy, width, height, verticesX, verticesY  );
                i2 = getNumVertices();
                createVertex(sx+1, sy+1, width, height, verticesX, verticesY);
                addTriangle(i0, i1, i2);

                i0 = getNumVertices();
                createVertex(sx  , sy, width, height, verticesX, verticesY  );
                i1 = getNumVertices();
                createVertex(sx+1, sy+1, width, height, verticesX, verticesY);
                i2 = getNumVertices();
                createVertex(sx  , sy+1, width, height, verticesX, verticesY);
                addTriangle(i0, i1, i2);
            }
        }
    }
    
    /**
     * Create a single vertex with texture coordinates and normal
     * 
     * @param sx Grid index X
     * @param sy Grid index Y
     * @param width The size in x direction
     * @param height The size in y direction
     * @param verticesX The number of grid vertices in x direction
     * @param verticesY The number of grid vertices in y direction 
     */
    private void createVertex(
        int sx, int sy, 
        int width, int height, 
        int verticesX, int verticesY)
    {
        float stepX = 1.0f / (verticesX - 1);
        float stepY = 1.0f / (verticesY - 1);
        float relX = sx * stepX;
        float relY = sy * stepY;

        float posX = -0.5f + relX;
        float posY = -0.5f + relY;
        Point3f vertex = new Point3f();
        vertex.x = posX * width;
        vertex.y = posY * height;
        vertex.z = 0;
        addVertex(vertex);

        TexCoord2f texCoord = new TexCoord2f();
        texCoord.x = relX;
        texCoord.y = 1 - relY;
        addTexCoord(texCoord);
        
        Vector3f normal = new Vector3f(0,0,1);
        addNormal(normal);
    }
    
}
