/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.interaction;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.JComponent;

import de.javagl.swogl.Condition;
import de.javagl.swogl.Conditions;
import de.javagl.swogl.IntersectionInfo;
import de.javagl.swogl.MouseEventConditions;
import de.javagl.swogl.PickingInfo;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglContainer;
import de.javagl.swogl.geometry.SimpleGeometry;

/**
 * Implementation of a default {@link Control} for a MagnifyingGlass.
 */
public class MagnifyingGlassControl implements Control
{
    /**
     * The listener of this Control, which will be attached to 
     * the component. 
     */
    private class Listener implements MouseMotionListener, MouseWheelListener
    {
        @Override
        public void mouseDragged(MouseEvent e)
        {
        }

        @Override
        public void mouseMoved(MouseEvent e)
        {
            if (!isEnabled())
            {
                return;
            }
            
            if (!focusingCondition.is(e))
            {
                return;
            }

            // Compute the picking information for a ray sent into the 
            // SwoglContainer at the current mouse position, and check 
            // if it hit the SwoglComponent
            IntersectionInfo intersectionInfo = 
                swoglContainer.computeIntersectionInfo(e.getPoint());
            if (intersectionInfo == null ||
                intersectionInfo.getSwoglComponent() != swoglComponent)
            {
                return;
            }

            // Set the position on the SwoglComponent that was hit with the
            // mouse as the new position for the magnifying glass
            PickingInfo pickingInfo = intersectionInfo.createPickingInfo();
            Point position = pickingInfo.getPoint();
            magnifyingGlass.setFocus(position.x, position.y);
            //System.out.println("Focus at "+position);
            
            magnifyingGlass.transformTexCoords(geometry);
            swoglContainer.repaint();
        }

        @Override
        public void mouseWheelMoved(MouseWheelEvent e)
        {
            if (!isEnabled())
            {
                return;
            }
            
            boolean affectMagnification = magnificationCondition.is(e);
            boolean affectRadius = radiusCondition.is(e);
            if (!affectMagnification && !affectRadius)
            {
                return;
            }
            
            IntersectionInfo intersectionInfo = 
                swoglContainer.computeIntersectionInfo(e.getPoint());
            if (intersectionInfo == null ||
                intersectionInfo.getSwoglComponent() != swoglComponent)
            {
                return;
            }

            // Change the radius of the magnifying glass
            if (affectRadius)
            {
                float radius = magnifyingGlass.getRadius();
                radius += 50 * e.getWheelRotation();
                magnifyingGlass.setRadius(radius);
                //System.out.println("radius "+radius);
            }
            
            // Change the magnification of the magnifying glass
            if (affectMagnification)
            {
                float magnification = magnifyingGlass.getMagnification();
                magnification += 0.01f * e.getWheelRotation();
                magnifyingGlass.setMagnification(magnification);
                //System.out.println("magnification "+magnification);
            }

            // Trigger an update of the texture coordinates of the 
            // geometry, which will then be transformed by the 
            // magnifying glass according to its current parameters
            magnifyingGlass.transformTexCoords(geometry);
            swoglContainer.repaint();
        }
    }

    /**
     * The SwoglContainer that contains the SwoglComponent that
     * is to be magnified
     */
    private SwoglContainer swoglContainer;

    /**
     * The SwoglComponent that is to be magnified
     */
    private SwoglComponent swoglComponent;
    
    /**
     * The Geometry of the SwoglComponent.
     */
    private SimpleGeometry geometry;
    
    /**
     * The MagnifyingGlass which modifies the texture coordinates 
     * of the geometry, to achieve the magnification effect.
     */
    private MagnifyingGlass magnifyingGlass;
    
    /**
     * The Listener instance
     */
    private final Listener listener = new Listener();

    /**
     * Whether this control is currently enabled
     */
    private boolean enabled = true;
    
    /**
     * The condition which states whether mouse events should 
     * cause the focus to be located at the mouse position
     */
    private Condition<MouseEvent> focusingCondition;
    
    /**
     * The condition which states whether mouse events should 
     * change the radius of the magnifying glass
     */
    private Condition<MouseEvent> radiusCondition;
    
    /**
     * The condition which states whether mouse events should 
     * change the magnification of the magnifying glass
     */
    private Condition<MouseEvent> magnificationCondition;
    
    
    /**
     * Creates a new Magnifier that magnifies the contents
     * of the given SwoglComponent, which resides in the
     * given SwoglContainer and contains the given Geometry.
     * 
     * @param swoglContainer The SwoglContainer that contains the 
     * SwoglComponent
     * @param swoglComponent The swoglComponent that is to be magnified
     * @param geometry The geometry that the SwoglComponent consists of
     */
    public MagnifyingGlassControl(
        SwoglContainer swoglContainer, 
        SwoglComponent swoglComponent,
        SimpleGeometry geometry)
    {
        this.swoglContainer = swoglContainer;
        this.swoglComponent = swoglComponent;
        this.geometry = geometry;
        
        Dimension size = swoglComponent.getComponent().getSize();
        magnifyingGlass = new MagnifyingGlass(size.width, size.height);
        magnifyingGlass.setMagnification(0.9f);
        magnifyingGlass.setRadius(size.width / 6);
        
        focusingCondition = MouseEventConditions.trueCondition();
        magnificationCondition = Conditions.not(MouseEventConditions.controlDownCondition());
        radiusCondition = MouseEventConditions.controlDownCondition();
    }

    
    @Override
    public void attachTo(JComponent component)
    {
        component.addMouseMotionListener(listener);
        component.addMouseWheelListener(listener);
    }

    @Override
    public void detachFrom(JComponent component)
    {
        component.removeMouseMotionListener(listener);
        component.removeMouseWheelListener(listener);
    }
    
    @Override
    public void setEnabled(boolean enabled)
    {
        this.enabled = enabled;
    }

    @Override
    public boolean isEnabled()
    {
        return enabled;
    }


    /**
     * Set the condition which states whether mouse events should 
     * change the focus location of the magnifying glass
     *
     * @param focusingCondition The focusing condition to set
     */
    public void setFocusingCondition(Condition<MouseEvent> focusingCondition)
    {
        this.focusingCondition = focusingCondition;
    }


    /**
     * Set the condition which states whether mouse events should 
     * change the radius of the magnifying glass
     *
     * @param radiusCondition The radius condition to set
     */
    public void setRadiusCondition(Condition<MouseEvent> radiusCondition)
    {
        this.radiusCondition = radiusCondition;
    }


    /**
     * Set the condition which states whether mouse events should 
     * change the magnification of the magnifying glass
     *
     * @param magnificationCondition The magnification condition to set
     */
    public void setMagnificationCondition(
        Condition<MouseEvent> magnificationCondition)
    {
        this.magnificationCondition = magnificationCondition;
    }
    
    
}