/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.interaction;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.JComponent;

import de.javagl.swogl.Condition;
import de.javagl.swogl.MouseEventConditions;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglContainer;

/**
 * Implementation of a {@link Control} that allows manipulating
 * {@link SwoglComponent SwoglComponents} inside a {@link SwoglContainer}.
 * It combines a {@link RotatingBehavior}, a {@link MovingBehavior}
 * and a {@link ScalingBehavior} and allows activating them based
 * on {@link Condition Conditions} that may be set for the incoming
 * mouse events.
 */
public class SpaceInteractionControl implements Control
{
    /**
     * A constant for the movement speed of SwoglComponents when
     * the mouse wheel is rotated
     */
    private static final float MOUSE_WHEEL_MOVEMENT_SPEED = 50.0f;

    
    /**
     * The listener class that will be attached to or detached from
     * the component.
     */
    private class Listener 
        implements MouseListener, MouseMotionListener, MouseWheelListener
    {
        @Override
        public void mouseWheelMoved(MouseWheelEvent e)
        {
            if (!enabled)
            {
                return;
            }
            movingBehavior.startMovement(e.getPoint());
            int deltaZ = 
                -(int)(e.getWheelRotation()*MOUSE_WHEEL_MOVEMENT_SPEED);
            movingBehavior.doMovement(e.getPoint(), deltaZ);
        }

        @Override
        public void mousePressed(MouseEvent e)
        {
            if (!enabled)
            {
                return;
            }
            if (rotatingCondition.is(e))
            {
                rotatingBehavior.startRotation(e.getPoint());
            }
            if (scalingCondition.is(e))
            {
                scalingBehavior.startScaling(e.getPoint());
            }
            if (movingCondition.is(e))
            {
                movingBehavior.startMovement(e.getPoint());
            }
        }
        
        @Override
        public void mouseDragged(MouseEvent e)
        {
            if (!enabled)
            {
                return;
            }
            if (rotatingCondition.is(e))
            {
                rotatingBehavior.doRotation(e.getPoint());
            }
            if (scalingCondition.is(e))
            {
                scalingBehavior.doScaling(e.getPoint());
            }
            if (movingCondition.is(e))
            {
                movingBehavior.doMovement(e.getPoint(), 0);
            }
            
        }

        @Override
        public void mouseMoved(MouseEvent e)
        {
        }

        @Override
        public void mouseClicked(MouseEvent e)
        {
        }

        @Override
        public void mouseReleased(MouseEvent e)
        {
        }

        @Override
        public void mouseEntered(MouseEvent e)
        {
        }

        @Override
        public void mouseExited(MouseEvent e)
        {
        }
        
    }
    
    /**
     * The rotating behavior
     */
    private RotatingBehavior rotatingBehavior;

    /**
     * The condition which states which mouse events should be
     * passed to the rotating behavior
     */
    private Condition<MouseEvent> rotatingCondition;
    
    /**
     * The moving behavior
     */
    private MovingBehavior movingBehavior;

    /**
     * The condition which states which mouse events should be
     * passed to the moving behavior
     */
    private Condition<MouseEvent> movingCondition;
    
    /**
     * The scaling behavior
     */
    private ScalingBehavior scalingBehavior;

    /**
     * The condition which states which mouse events should be
     * passed to the scaling behavior
     */
    private Condition<MouseEvent> scalingCondition;
    
    
    /**
     * The Listener instance
     */
    private final Listener listener = new Listener();

    /**
     * Whether this control is currently enabled
     */
    private boolean enabled = true;
    
    /**
     * Creates a new space interaction control for the given SwoglContainer.
     * The default conditions for the behaviors will be
     * <ul>
     *    <li>Mouse button 1 for rotation</li>
     *    <li>Mouse button 2 for scaling</li>
     *    <li>Mouse button 3 for moving</li>
     * </ul>
     * The mouse wheel will move the SwoglComponents along the Z-axis
     * 
     * @param swoglContainer The SwoglContainer to operate on
     */
    public SpaceInteractionControl(SwoglContainer swoglContainer)
    {
        rotatingBehavior = new RotatingBehavior(swoglContainer);
        rotatingCondition = MouseEventConditions.buttonCondition(1);
        
        movingBehavior = new MovingBehavior(swoglContainer);
        movingCondition = MouseEventConditions.buttonCondition(3);
        
        scalingBehavior = new ScalingBehavior(swoglContainer);
        scalingCondition = MouseEventConditions.buttonCondition(2);
    }
    
    /**
     * Set the condition for passing events to the {@link RotatingBehavior}
     * 
     * @param rotatingCondition The rotating condition to set
     */
    public void setRotatingCondition(Condition<MouseEvent> rotatingCondition)
    {
        this.rotatingCondition = rotatingCondition;
    }

    /**
     * Set the condition for passing events to the {@link ScalingBehavior}
     * 
     * @param scalingCondition The scaling condition to set
     */
    public void setScalingCondition(Condition<MouseEvent> scalingCondition)
    {
        this.scalingCondition = scalingCondition;
    }

    /**
     * Set the condition for passing events to the {@link MovingBehavior}
     * 
     * @param movingCondition The moving condition to set
     */
    public void setMovingCondition(Condition<MouseEvent> movingCondition)
    {
        this.movingCondition = movingCondition;
    }
    
    @Override
    public void attachTo(JComponent component)
    {
        component.addMouseListener(listener);
        component.addMouseMotionListener(listener);
        component.addMouseWheelListener(listener);
    }

    @Override
    public void detachFrom(JComponent component)
    {
        component.removeMouseListener(listener);
        component.removeMouseMotionListener(listener);
        component.removeMouseWheelListener(listener);
    }
    
    @Override
    public void setEnabled(boolean enabled)
    {
        this.enabled = enabled;
    }

    @Override
    public boolean isEnabled()
    {
        return enabled;
    }


}
