/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.layout;

import java.util.ArrayList;
import java.util.List;

import javax.vecmath.Matrix4f;

import de.javagl.swogl.MatrixUtils;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglContainer;
import de.javagl.swogl.geometry.Box;
import de.javagl.swogl.geometry.Geometry;
import de.javagl.swogl.geometry.GeometryUtils;

/**
 * An implementation of the LayoutManager3D interface that arranges
 * the components as a wheel. 
 */
public class WheelLayout3D implements LayoutManager3D
{

    /**
     * Time in milliseconds between the animation steps
     */
    private static final int ROTATION_DELAY = 30;

    /**
     * The duration of a single step of the wheel rotation
     */
    private int stepDurationMS = 250;
    
    /**
     * The rotation angle between two components
     */
    private float angleDelta = 0;

    /**
     * The current rotation angle of the wheel
     */
    private float currentAngleOffset = 0;
    
    /**
     * The rotation angle the wheel animation is heading for
     */
    private float targetAngleOffset = 0;

    /**
     * The WheelAnimator that animates the wheel
     */
    private WheelAnimator wheelAnimator;
    
    /**
     * The SwoglContainer this layout belongs to
     */
    private SwoglContainer swoglContainer;
    
    /**
     * Creates a new WheelLayout3D
     */
    public WheelLayout3D()
    {
    }

    /**
     * Set the duration of a single step of the wheel rotation
     * in milliseconds (this is the time for the transition
     * between two components)
     * 
     * @param ms The duration in milliseconds
     */
    public void setStepDuration(int ms)
    {
        this.stepDurationMS = ms;
    }

    @Override
    public void setSwoglContainer(SwoglContainer swoglContainer)
    {
        this.swoglContainer = swoglContainer;
        if (wheelAnimator != null)
        {
            wheelAnimator.stop();
            wheelAnimator = null;
        }
        if (swoglContainer != null)
        {
            wheelAnimator = new WheelAnimator();
            Thread animatorThread = new Thread(wheelAnimator);
            animatorThread.setDaemon(true);
            animatorThread.start();
        }
    }

    @Override
    public void doLayout3D()
    {
        float radius = 0.0f;
        angleDelta = 0;
        int componentCount = swoglContainer.getSwoglComponentCount();
        if (componentCount > 1)
        {
        	// Compute the height of the highest SwoglComponent
        	float maxHeight = Float.NEGATIVE_INFINITY;
        	List<Box> boundingBoxes = new ArrayList<Box>();
        	for (int i = 0; i < componentCount; i++)
        	{
        		SwoglComponent swoglComponent =
        		    swoglContainer.getSwoglComponent(i);
                Geometry geometry = swoglComponent.getGeometry();
                Box boundingBox = 
                    GeometryUtils.computeBoundingBox(geometry);
        		boundingBoxes.add(boundingBox);
        		maxHeight = Math.max(maxHeight, boundingBox.getSizeY());
        	}

        	// Compute the angle between the SwoglComponents and
        	// the wheel radius, and the rotation speed
        	angleDelta = (float) (Math.PI * 2) / componentCount;
        	radius = maxHeight / (2 * (float) Math.tan(angleDelta / 2));
        }

        // Arrange all SwoglComponents in a wheel shape
        for (int i = 0; i < componentCount; i++)
        {
        	SwoglComponent swoglComponent = 
        	    swoglContainer.getSwoglComponent(i);
            Geometry geometry = swoglComponent.getGeometry();
            Box boundingBox = 
                GeometryUtils.computeBoundingBox(geometry);
            Matrix4f transform = MatrixUtils.mul(
        	    MatrixUtils.translate(0, 0, -radius),
        	    MatrixUtils.rotateX(i * angleDelta + currentAngleOffset),
        	    MatrixUtils.translate(0, 0, 
        	        radius + boundingBox.getSizeZ() + 0.1f));
        	swoglComponent.setTransform(transform);
        }
    }

    /**
     * Rotate the wheel about the given number of steps
     * 
     * @param steps The number of steps to rotate
     */
    public void rotateWheel(int steps)
    {
        while (currentAngleOffset > Math.PI * 2)
        {
            currentAngleOffset -= Math.PI * 2;
        }
        
        // Compute the target angle from the current angle
        // and the number of steps that should be rotated.
        int currentStep = Math.round(currentAngleOffset / angleDelta);
        int targetStep = (currentStep + steps);
        targetAngleOffset = targetStep * angleDelta;

        //System.out.println("rotation " + steps);
        //System.out.println("current " + currentStep + " target " + targetStep);
        //System.out.println("currentAngle " + currentAngleOffset + " targetAngle " + targetAngleOffset);

        synchronized(this)
        {
            // Notify the WheelAnimator
            notifyAll();
        }
        
    }

    /**
     * The class that performs the wheel rotation animation
     */
    class WheelAnimator implements Runnable
    {
        private boolean stopped = false;
        
        void stop()
        {
            synchronized (WheelLayout3D.this)
            {
                stopped = true;
                WheelLayout3D.this.notifyAll();
            }
        }
        
        @Override
        public void run()
        {
            while (!stopped)
            {
                float rotationAngleStepSize = 
                    angleDelta / (stepDurationMS / ROTATION_DELAY);

                if (currentAngleOffset != targetAngleOffset)
                {
                    // Head toward the targetAngleOffset
                    if (currentAngleOffset < targetAngleOffset)
                    {
                        float a = currentAngleOffset + rotationAngleStepSize;
                        currentAngleOffset = Math.min(a, targetAngleOffset);
                    }
                    else if (currentAngleOffset > targetAngleOffset)
                    {
                        float a = currentAngleOffset - rotationAngleStepSize;
                        currentAngleOffset = Math.max(a, targetAngleOffset);
                    }
                    
                    swoglContainer.validate3D();
                    try
                    {
                        Thread.sleep(ROTATION_DELAY);
                    }
                    catch (InterruptedException e)
                    {
                        Thread.currentThread().interrupt();
                    }
                }
                else
                {
                    // Wait until this WheelAnimator is triggered by
                    // rotating the mouse wheel
                    synchronized (WheelLayout3D.this)
                    {
                        try
                        {
                            WheelLayout3D.this.wait();
                        }
                        catch (InterruptedException e)
                        {
                            Thread.currentThread().interrupt();
                        }
                    }
                }
            }
        }
    }
    

}
