/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2011 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import java.util.Collections;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglRenderer;
import de.javagl.swogl.Texture;
import de.javagl.swogl.geometry.Geometry;

/**
 * Abstract base implementation of a {@link SwoglRenderer}. This class 
 * provides facilities to maintain a set of 
 * {@link SwoglComponent SwoglComponents} and to initialize, update
 * or delete these components on the rendering thread. These tasks
 * are delegated to protected methods that may be implemented 
 * accordingly.
 */
abstract class AbstractSwoglRenderer implements SwoglRenderer
{
    /**
     * Compile-time flag which indicates whether the stack
     * traces of runnables should be preserved and printed 
     * when a runnable causes an exception.
     */
    private static final boolean PRESERVE_STACK_TRACES = true;

    /**
     * The SwoglComponents that are contained in renderer
     */
    private final List<SwoglComponent> swoglComponents;
    
    /**
     * The SwoglComponents for which an internal representation
     * already has been created.
     */
    private final List<SwoglComponent> handledSwoglComponents;
    
    /**
     * The tasks that have to be executed before the next rendering
     */
    private final List<Runnable> tasks;
    
    /**
     * Creates a new AbstractSwoglRenderer
     */
    protected AbstractSwoglRenderer()
    {
        swoglComponents = new CopyOnWriteArrayList<SwoglComponent>();
        handledSwoglComponents = new CopyOnWriteArrayList<SwoglComponent>();
        tasks = new CopyOnWriteArrayList<Runnable>();
    }

    @Override
    public void handleSwoglComponent(final SwoglComponent swoglComponent)
    {
        swoglComponents.add(swoglComponent);
        tasks.add(wrap(new Runnable()
        {
            @Override
            public void run()
            {
                // Only handle the SwoglComponent if it has not
                // already been removed again
                if (swoglComponents.contains(swoglComponent))
                {
                    handleSwoglComponentInternal(swoglComponent);
                    handledSwoglComponents.add(swoglComponent);
                }
            }
        }));
    }
    
    /**
     * Will be called on the rendering thread to handle the given 
     * SwoglComponent.
     * 
     * @param swoglComponent The SwoglComponent to handle
     */
    protected abstract void handleSwoglComponentInternal(
        SwoglComponent swoglComponent);

    @Override
    public void releaseSwoglComponent(final SwoglComponent swoglComponent)
    {
        swoglComponents.remove(swoglComponent);
        tasks.add(wrap(new Runnable()
        {
            @Override
            public void run()
            {
                if (handledSwoglComponents.contains(swoglComponent))
                {
                    releaseSwoglComponentInternal(swoglComponent);
                    handledSwoglComponents.remove(swoglComponent);
                }
            }
        }));
    }
    
    /**
     * Will be called on the rendering thread to release the given 
     * SwoglComponent.
     * 
     * @param swoglComponent The SwoglComponent to release
     */
    protected abstract void releaseSwoglComponentInternal(
        SwoglComponent swoglComponent);
    

    @Override 
    public List<SwoglComponent> getSwoglComponents()
    {
        return Collections.unmodifiableList(swoglComponents);
    }
    
    @Override
    public void renderAllSwoglComponents()
    {
        for (SwoglComponent swoglComponent : swoglComponents)
        {
            renderSwoglComponent(swoglComponent);
        }
    }

    /**
     * Prepare rendering by executing all tasks that have been
     * scheduled for the rendering thread.
     */
    private void prepareRender()
    {
        for (Runnable task : tasks)
        {
            task.run();
        }
        tasks.clear();
    }
    
    @Override
    public void renderSwoglComponent(SwoglComponent swoglComponent)
    {
        prepareRender();
        if (!handledSwoglComponents.contains(swoglComponent))
        {
            throw new IllegalArgumentException(
                "SwoglComponent is not handled by this renderer");
        }
        updateSwoglComponentInternal(swoglComponent);
        renderSwoglComponentInternal(swoglComponent);
    }

    /**
     * Update the state of the given SwoglComponent on the rendering
     * thread. By default, this will check the dirty flags of the
     * geometry and texture, call the 
     * {@link #updateTexCoordsInternal(Geometry)},
     * {@link #updateVerticesInternal(Geometry)},
     * {@link #updateTextureInternal(Texture)},
     * methods if necessary, and clear the flags accordingly.
     * 
     * @param swoglComponent The SwoglComponent to update.
     */
    protected void updateSwoglComponentInternal(SwoglComponent swoglComponent)
    {
        Geometry geometry = swoglComponent.getGeometry();
        Texture texture = swoglComponent.getTexture();

        if (geometry.areTexCoordsModified())
        {
            updateTexCoordsInternal(geometry);
            geometry.setTexCoordsModified(false);
        }
        if (geometry.areVerticesModified())
        {
            updateVerticesInternal(geometry);
            geometry.setVerticesModified(false);
        }
        if (texture.getDirtyRectangle() != null)
        {
            updateTextureInternal(texture);
            texture.clearDirtyRectangle();
        }
    }
    
    /**
     * Update the texture coordinates of the given geometry.
     * 
     * @param geometry The geometry to update.
     */
    protected void updateTexCoordsInternal(Geometry geometry)
    {
        // Empty, may be overridden
    }
    
    /**
     * Update the vertices and normals of the given geometry.
     * 
     * @param geometry The geometry to update.
     */
    protected void updateVerticesInternal(Geometry geometry)
    {
        // Empty, may be overridden
    }

    /**
     * Update the given texture.
     * 
     * @param texture The texture to update.
     */
    protected void updateTextureInternal(Texture texture)
    {
        // Empty, may be overridden
    }
    
    /**
     * Called on the rendering thread to actually render the given
     * SwoglComponent.
     * 
     * @param swoglComponent The SwoglComponent to render
     */
    protected abstract void renderSwoglComponentInternal(
        SwoglComponent swoglComponent);

    
    
    
    
    
    
    
    /**
     * When {@link #PRESERVE_STACK_TRACES} is <code>true</code>, this method 
     * wraps the given Runnable into one that preserves the stack trace of 
     * this method call, and, if the command causes an exception, prints 
     * it together with the stack trace of the exception caused by the 
     * command. <br /> 
     * <br />
     * When {@link #PRESERVE_STACK_TRACES} is <code>false</code>, the given
     * command is returned.
     * 
     * @param command The command to wrap.
     * @return The wrapped command or the command
     */
    private static Runnable wrap(final Runnable command)
    {
        if (!PRESERVE_STACK_TRACES)
        {
            return command;
        }
        
        final StackTraceElement ste[] = Thread.currentThread().getStackTrace();
        return new Runnable()
        {
            @Override
            public void run()
            {
                try
                {
                    command.run();
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                    System.err.println("Scheduled at");
                    for (StackTraceElement s : ste)
                    {
                        System.err.println("\tat "+s);
                    }
                }
            }
        };
    }
    
}
